pub mod positioncoloruv;
pub mod positionnormal;
pub mod positionnormalbone;
pub mod positionnormalcolor;
pub mod positionnormalcoloruv;
pub mod positionnormalcoloruvbone;
pub mod positionnormalcoloruvnormaluv;
pub mod positionnormalcoloruvnormaluvbone;
pub mod positiononly;

pub mod prelude;

use crate::{
    asset::MatrixBuffer, rasterizer_descriptor_layouts::RasterizerDescriptorLayouts,
    raytracer_descriptor_layouts::RaytracerDesciptorLayouts, scene_type::SceneType, Asset,
};

use utilities::{impl_reprc, prelude::*};
use vulkan_rs::prelude::*;

use gltf_loader::prelude::*;

use positioncoloruv::PositionColorUV;
use positionnormal::PositionNormal;
use positionnormalbone::PositionNormalBone;
use positionnormalcolor::PositionNormalColor;
use positionnormalcoloruv::PositionNormalColorUV;
use positionnormalcoloruvbone::PositionNormalColorUVBone;
use positionnormalcoloruvnormaluv::PositionNormalColorUVNormalUV;
use positionnormalcoloruvnormaluvbone::PositionNormalColorUVNormalUVBone;
use positiononly::PositionOnly;

use std::sync::Arc;

use anyhow::Result;

impl_reprc!(
    #[derive(Copy, Debug)]
    pub struct Material {
        #[assume_reprc]
        pub color: [f32; 4],

        #[assume_reprc]
        pub color_texture_index: i32,
        #[assume_reprc]
        pub normal_texture_index: i32,
        #[assume_reprc]
        pub emissive_texture_index: i32,
        #[assume_reprc]
        pub roughness_texture_index: i32,

        #[assume_reprc]
        pub emissive_factor: [f32; 4],

        #[assume_reprc]
        pub metallic_factor: f32,
        #[assume_reprc]
        pub roughness_factor: f32,
        #[assume_reprc]
        pub alpha_mode: u32,
        #[assume_reprc]
        pub alpha_cut_off: f32,
    }
);

impl From<PrimitiveMaterial> for Material {
    fn from(mat: PrimitiveMaterial) -> Self {
        Material {
            color: mat.color,

            color_texture_index: -1,
            normal_texture_index: -1,
            emissive_texture_index: -1,
            roughness_texture_index: -1,

            metallic_factor: mat.metallic_factor,
            emissive_factor: [
                mat.emissive_factor[0],
                mat.emissive_factor[1],
                mat.emissive_factor[2],
                0.0,
            ],

            roughness_factor: mat.roughness_factor,
            alpha_mode: match mat.alpha_mode {
                AlphaMode::Opaque => 0,
                AlphaMode::Mask => 1,
                AlphaMode::Blend => 2,
            },
            alpha_cut_off: mat.alpha_cut_off,
        }
    }
}

pub trait CreatePrimitiveData<T: ReprC + Copy + Send + Sync + 'static>: Sized {
    #[allow(clippy::too_many_arguments)]
    fn new(
        device: &Arc<Device>,
        vertex_buffer: Arc<Buffer<T>>,
        color_texture: Option<Arc<Image>>,
        normal_texture: Option<Arc<Image>>,
        model_buffer: &Arc<Buffer<MatrixBuffer>>,
        bone_buffer: Option<&Arc<Buffer<MatrixBuffer>>>,
        scene_type: SceneType,
        material: PrimitiveMaterial,
        opaque: bool,
    ) -> Result<Self>;
}

macro_rules! GeneratePrimitiveDataTypes {
    ([$($name: ident $(,)? )+]) => {
    #[derive(Clone, Debug)]
        pub enum PrimitiveDataTypes {
        $(
            $name(PrimitiveData<$name>),
        )+
        }

        impl PrimitiveDataTypes {
            pub fn clone(
                &self,
                model_buffer: &Arc<Buffer<MatrixBuffer>>,
                bone_buffer: Option<&Arc<Buffer<MatrixBuffer>>>,
            ) -> Result<PrimitiveDataTypes> {
                match self {
                    $(
                        Self::$name(data) => Ok(Self::$name(data.reuse(model_buffer, bone_buffer)?)),
                    )+
                }
            }

            pub fn vertex_count(&self) -> u32 {
                match self {
                    $(
                        Self::$name(data) => data.vertex_buffer.size() as u32,
                    )+
                }
            }

            pub fn descriptor_set(&self) -> &Arc<DescriptorSet> {
                match self {
                    $(
                        Self::$name(data) => &data.descriptor_set,
                    )+
                }
            }

            pub fn is_opaque(&self) -> bool {
                match self {
                    $(
                        Self::$name(data) => data.opaque,
                    )+
                }
            }

            pub fn material_buffer(&self) -> &Arc<Buffer<Material>> {
                match self {
                    $(
                        Self::$name(data) => &data.material_buffer,
                    )+
                }
            }

            pub fn color_texture(&self) -> &Option<Arc<Image>> {
                match self {
                    $(
                        Self::$name(data) => &data.color_texture,
                    )+
                }
            }

            pub fn update_color_image(&mut self, image: Option<Arc<Image>>) -> Result<()> {
                match self {
                    $(
                        Self::$name(data) => data.update_color_image(image),
                    )+
                }
            }
        }

        $(
            impl CreatePrimitiveData<$name> for PrimitiveDataTypes {
                fn new(
                    device: &Arc<Device>,
                    vertex_buffer: Arc<Buffer<$name>>,
                    color_texture: Option<Arc<Image>>,
                    normal_texture: Option<Arc<Image>>,
                    model_buffer: &Arc<Buffer<MatrixBuffer>>,
                    bone_buffer: Option<&Arc<Buffer<MatrixBuffer>>>,
                    scene_type: SceneType,
                    material: PrimitiveMaterial,
                    opaque: bool,
                ) -> Result<Self> {
                    Ok(Self::$name(PrimitiveData::new(
                        device,
                        vertex_buffer,
                        color_texture,
                        normal_texture,
                        model_buffer,
                        bone_buffer,
                        scene_type,
                        material,
                        opaque,
                    )?))
                }
            }
        )+
    };
}

GeneratePrimitiveDataTypes!([
    PositionOnly,
    PositionNormal,
    PositionNormalColor,
    PositionNormalBone,
    PositionColorUV,
    PositionNormalColorUV,
    PositionNormalColorUVNormalUV,
    PositionNormalColorUVBone,
    PositionNormalColorUVNormalUVBone
]);

#[derive(Clone, Debug)]
pub struct PrimitiveData<T: ReprC + Copy> {
    pub vertex_buffer: Arc<Buffer<T>>,
    pub color_texture: Option<Arc<Image>>,
    pub normal_texture: Option<Arc<Image>>,
    pub descriptor_set: Arc<DescriptorSet>,
    pub material_buffer: Arc<Buffer<Material>>,
    pub opaque: bool,

    device: Arc<Device>,
    render_type: SceneType,
}

impl PrimitiveData<PositionColorUV> {
    pub fn remove_uv_from_buffer(&self) -> Result<Arc<Buffer<PositionOnly>>> {
        Asset::create_buffer(
            &self.device,
            &None,
            self.vertex_buffer
                .map_complete()?
                .iter()
                .map(|vertex| PositionOnly::new(vertex.position))
                .collect::<Vec<_>>(),
        )
    }
}

impl PrimitiveData<PositionNormalColorUV> {
    pub fn remove_uv_from_buffer(&self) -> Result<Arc<Buffer<PositionNormal>>> {
        Asset::create_buffer(
            &self.device,
            &None,
            self.vertex_buffer
                .map_complete()?
                .iter()
                .map(|vertex| PositionNormal::new(vertex.position, vertex.normal))
                .collect::<Vec<_>>(),
        )
    }
}

impl PrimitiveData<PositionNormalColorUVNormalUV> {
    pub fn remove_uv_from_buffer(&self) -> Result<Arc<Buffer<PositionNormal>>> {
        Asset::create_buffer(
            &self.device,
            &None,
            self.vertex_buffer
                .map_complete()?
                .iter()
                .map(|vertex| PositionNormal::new(vertex.position, vertex.normal))
                .collect::<Vec<_>>(),
        )
    }
}

impl PrimitiveData<PositionNormalColorUVBone> {
    pub fn remove_uv_from_buffer(&self) -> Result<Arc<Buffer<PositionNormalBone>>> {
        Asset::create_buffer(
            &self.device,
            &None,
            self.vertex_buffer
                .map_complete()?
                .iter()
                .map(|vertex| {
                    PositionNormalBone::new(
                        vertex.position,
                        vertex.normal,
                        vertex.joints,
                        vertex.weights,
                    )
                })
                .collect::<Vec<_>>(),
        )
    }
}

impl PrimitiveData<PositionNormalColorUVNormalUVBone> {
    pub fn remove_uv_from_buffer(&self) -> Result<Arc<Buffer<PositionNormalBone>>> {
        Asset::create_buffer(
            &self.device,
            &None,
            self.vertex_buffer
                .map_complete()?
                .iter()
                .map(|vertex| {
                    PositionNormalBone::new(
                        vertex.position,
                        vertex.normal,
                        vertex.joints,
                        vertex.weights,
                    )
                })
                .collect::<Vec<_>>(),
        )
    }
}

impl<T: ReprC + Copy + Send + Sync + 'static> PrimitiveData<T> {
    #[allow(clippy::too_many_arguments)]
    fn new(
        device: &Arc<Device>,
        vertex_buffer: Arc<Buffer<T>>,
        color_texture: Option<Arc<Image>>,
        normal_texture: Option<Arc<Image>>,
        model_buffer: &Arc<Buffer<MatrixBuffer>>,
        bone_buffer: Option<&Arc<Buffer<MatrixBuffer>>>,
        render_type: SceneType,
        material: PrimitiveMaterial,
        opaque: bool,
    ) -> Result<PrimitiveData<T>> {
        let material_buffer = Buffer::builder()
            .set_usage(VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT)
            .set_memory_usage(MemoryUsage::CpuOnly)
            .set_data(&[Material::from(material)])
            .build(device.clone())?;

        let descriptor_set = Self::create_descriptor_set(
            device,
            &vertex_buffer,
            &color_texture,
            &normal_texture,
            model_buffer,
            bone_buffer,
            &material_buffer,
            render_type,
        )?;

        Ok(PrimitiveData {
            vertex_buffer,
            color_texture,
            normal_texture,
            descriptor_set,
            material_buffer,
            opaque,

            device: device.clone(),
            render_type,
        })
    }

    pub fn update_color_image(&mut self, color_image: Option<Arc<Image>>) -> Result<()> {
        // replace color texture handle
        self.color_texture = color_image;

        // update descriptor set
        if let SceneType::Rasterizer = self.render_type {
            if let Some(ref color) = self.color_texture {
                self.descriptor_set
                    .update(&[DescriptorWrite::combined_samplers(2, &[color])])?;
            }
        }

        Ok(())
    }

    pub fn update_normal_image(&mut self, normal_image: Option<Arc<Image>>) -> Result<()> {
        // replace normal texture handle
        self.normal_texture = normal_image;

        // update descriptor set
        if let SceneType::Rasterizer = self.render_type {
            if let Some(ref normal) = self.normal_texture {
                self.descriptor_set
                    .update(&[DescriptorWrite::combined_samplers(3, &[normal])])?;
            }
        }

        Ok(())
    }

    pub fn reuse(
        &self,
        model_buffer: &Arc<Buffer<MatrixBuffer>>,
        bone_buffer: Option<&Arc<Buffer<MatrixBuffer>>>,
    ) -> Result<PrimitiveData<T>> {
        let descriptor_set = Self::create_descriptor_set(
            &self.device,
            &self.vertex_buffer,
            &self.color_texture,
            &self.normal_texture,
            model_buffer,
            bone_buffer,
            &self.material_buffer,
            self.render_type,
        )?;

        Ok(PrimitiveData {
            vertex_buffer: self.vertex_buffer.clone(),
            color_texture: self.color_texture.clone(),
            normal_texture: self.normal_texture.clone(),
            descriptor_set,
            material_buffer: self.material_buffer.clone(),
            opaque: self.opaque,

            device: self.device.clone(),
            render_type: self.render_type,
        })
    }

    #[allow(clippy::too_many_arguments)]
    fn create_descriptor_set(
        device: &Arc<Device>,
        vertex_buffer: &Arc<Buffer<T>>,
        color_texture: &Option<Arc<Image>>,
        normal_texture: &Option<Arc<Image>>,
        model_buffer: &Arc<Buffer<MatrixBuffer>>,
        bone_buffer: Option<&Arc<Buffer<MatrixBuffer>>>,
        material_buffer: &Arc<Buffer<Material>>,
        render_type: SceneType,
    ) -> Result<Arc<DescriptorSet>> {
        match render_type {
            SceneType::Rasterizer => {
                let mut writes = vec![
                    DescriptorWrite::uniform_buffers(0, &[model_buffer]),
                    DescriptorWrite::uniform_buffers(1, &[material_buffer]),
                ];

                let descriptor_layout = match (color_texture, normal_texture, bone_buffer) {
                    (Some(color), Some(normal), Some(bone)) => {
                        writes.push(DescriptorWrite::combined_samplers(2, &[color]));
                        writes.push(DescriptorWrite::uniform_buffers(3, &[bone]));
                        writes.push(DescriptorWrite::combined_samplers(4, &[normal]));

                        RasterizerDescriptorLayouts::color_bone_normal_descriptor_layout(
                            device.clone(),
                        )
                    }
                    (Some(color), Some(normal), None) => {
                        writes.push(DescriptorWrite::combined_samplers(2, &[color]));
                        writes.push(DescriptorWrite::combined_samplers(4, &[normal]));

                        RasterizerDescriptorLayouts::color_normal_descriptor_layout(device.clone())
                    }
                    (Some(color), None, Some(bone)) => {
                        writes.push(DescriptorWrite::combined_samplers(2, &[color]));
                        writes.push(DescriptorWrite::uniform_buffers(3, &[bone]));

                        RasterizerDescriptorLayouts::color_bone_descriptor_layout(device.clone())
                    }
                    (Some(color), None, None) => {
                        writes.push(DescriptorWrite::combined_samplers(2, &[color]));

                        RasterizerDescriptorLayouts::color_descriptor_layout(device.clone())
                    }
                    (None, None, Some(bone)) => {
                        writes.push(DescriptorWrite::uniform_buffers(3, &[bone]));

                        RasterizerDescriptorLayouts::bone_descriptor_layout(device.clone())
                    }
                    (None, _, _) => RasterizerDescriptorLayouts::descriptor_layout(device.clone()),
                }?;

                let descriptor_set = DescriptorPool::builder()
                    .set_layout(descriptor_layout)
                    .build(device.clone())?
                    .prepare_set()
                    .allocate()?;

                descriptor_set.update(&writes)?;

                Ok(descriptor_set)
            }
            SceneType::Raytracer => {
                let descriptor_layout = match bone_buffer {
                    Some(_bones) => {
                        RaytracerDesciptorLayouts::bone_descriptor_layout(device.clone())?
                    }
                    None => RaytracerDesciptorLayouts::descriptor_layout(device.clone())?,
                };

                let descriptor_set = DescriptorPool::builder()
                    .set_layout(descriptor_layout)
                    .build(device.clone())?
                    .prepare_set()
                    .allocate()?;

                let mut writes = vec![
                    DescriptorWrite::storage_buffers(0, &[vertex_buffer]),
                    DescriptorWrite::uniform_buffers(1, &[model_buffer]),
                ];

                if let Some(bone) = bone_buffer {
                    writes.push(DescriptorWrite::uniform_buffers(2, &[bone]));
                }

                descriptor_set.update(&writes)?;

                Ok(descriptor_set)
            }
        }
    }
}
